//
// algebra.h
//
// Class and method declarations to do 2D and 3D linear algebra.
//
// Originally written by Jean-Francois DOUE, October 1993
// Modified by Craig Kaplan and Daniel Wood, April 1999

#ifndef __ALGEBRA_H__
#define __ALGEBRA_H__

#include <iostream>
#include <cmath>
#include <algorithm>

using namespace std;

class vec3;
class vec4;
class mat3;
class mat4;

// used as an exception during matrix inversion.
class SingularMatrix
{};

inline float minimum( float a, float b )
{
	return a < b ? a : b;
}

inline float maximum( float a, float b )
{
	return a > b ? a : b;
}

inline double minimum( double a, double b )
{
	return a < b ? a : b;
}

inline double maximum( double a, double b )
{
	return a > b ? a : b;
}

class vec3
{
public:
	// Constructors

	vec3() { n[0] = 0.0; n[1] = 0.0; n[2] = 0.0; }
	vec3( const double x, const double y, const double z )
		{ n[0] = x; n[1] = y; n[2] = z; }
//	vec3( const double d )
//		{ n[0] = d; n[1] = d; n[2] = d; }
	vec3( const vec3& v )
		{ n[0] = v.n[0]; n[1] = v.n[1]; n[2] = v.n[2]; }
	vec3( const vec4& v4 );

	vec3& operator	=( const vec3& v )
		{ n[0] = v.n[0]; n[1] = v.n[1]; n[2] = v.n[2]; return *this; }
	vec3& operator +=( const vec3& v )
		{ n[0] += v.n[0]; n[1] += v.n[1]; n[2] += v.n[2]; return *this; }
	vec3& operator -= ( const vec3& v )
		{ n[0] -= v.n[0]; n[1] -= v.n[1]; n[2] -= v.n[2]; return *this; }
	vec3& operator *= ( const double d )
		{ n[0] *= d; n[1] *= d; n[2] *= d; return *this; }
	vec3& operator /= ( const double d )
		{ n[0] /= d; n[1] /= d; n[2] /= d; return *this; }
	double& operator []( int i )
		{ return n[i]; }
	double operator []( int i ) const 
		{ return n[i]; }

	double maximum() const 
		{ return ::maximum( ::maximum( n[0], n[1] ), n[2] ); }

	// special functions

	double length2() const
		{ return n[0]*n[0] + n[1]*n[1] + n[2]*n[2]; }
	double length() const
		{ return sqrt( length2() ); }
	vec3 normalize() const
	{ 
		vec3 ret( *this );
		ret /= length();
		return ret;
	}	
	//	return *this / length(); }
#if 0
	friend vec3 operator -(const vec3& v);
	friend vec3 operator +(const vec3& a, const vec3& b);
	friend vec3 operator -(const vec3& a, const vec3& b);
	friend vec3 operator *(const vec3& a, const double d );
	friend vec3 operator *(const double d, const vec3& a);
	friend vec3 operator *(const mat4& a, const vec3& v);
	friend vec3 operator *(const vec3& v, mat4& a);
	friend double operator *(const vec3& a, const vec3& b);
	friend vec3 operator /(const vec3& a, const double d);
	friend vec3 operator ^(const vec3& a, const vec3& b);
	friend bool operator ==(const vec3& a, const vec3& b);
	friend bool operator !=(const vec3& a, const vec3& b);
	friend ostream& operator <<(ostream& s, const vec3& v);
	friend istream& operator >>(istream& s, vec3& v);
	friend void swap(vec3& a, vec3& b);	
	friend vec3 minimum(const vec3& a, const vec3& b);
	friend vec3 maximum(const vec3& a, const vec3& b);
	friend vec3 prod(const vec3& a, const vec3& b);
#endif
	// necessary friend declarations

	friend class mat3;
	friend class mat4;

	friend mat3 operator * (mat3& a, mat3& b);		    // matrix 3 product

public:
	double n[3];
};

class vec4
{
public:
	// Constructors

	vec4() { n[0] = 0.0; n[1] = 0.0; n[2] = 0.0; n[3] = 0.0; }
	vec4( const double x, const double y, const double z, const double w )
		{ n[0] = x; n[1] = y; n[2] = z; n[3] = w; }
//	vec4( const double d )
//		{ n[0] = d; n[1] = d; n[2] = d; n[3] = d; }
	vec4( const vec4& v )
		{ n[0] = v.n[0]; n[1] = v.n[1]; n[2] = v.n[2]; n[3] = v.n[3]; }
	vec4( const vec3& v )
		{ n[0] = v[0]; n[1] = v[1]; n[2] = v[2]; n[3] = 1.0; }

	vec4& operator =( const vec4& v )
		{ n[0] = v.n[0]; n[1] = v.n[1]; n[2] = v.n[2]; n[3] = v.n[3];
		  return *this; }
	vec4& operator +=( const vec4& v )
		{ n[0] += v.n[0]; n[1] += v.n[1]; n[2] += v.n[2]; n[3] += v.n[3];
		  return *this; }
	vec4& operator -= ( const vec4& v )
		{ n[0] -= v.n[0]; n[1] -= v.n[1]; n[2] -= v.n[2]; n[3] -= v.n[3];
		  return *this; }
	vec4& operator *= ( const double d )
		{ n[0] *= d; n[1] *= d; n[2] *= d; n[3] *= d; return *this; }
	vec4& operator /= ( const double d )
		{ n[0] /= d; n[1] /= d; n[2] /= d; n[3] /= d; return *this; }
	double& operator []( int i )
		{ return n[i]; }
	double operator []( int i ) const 
		{ return n[i]; }

	// special functions

	double length2() const
		{ return n[0]*n[0] + n[1]*n[1] + n[2]*n[2] + n[3]*n[3]; }
	double length() const
		{ return sqrt( length2() ); }
	vec4 normalize() const
		// { return *this / length(); }
	{ 
		vec4 ret( *this );
		ret /= length();
		return ret;
	}
#if 0
	friend vec4 operator -(const vec4& v);
	friend vec4 operator +(const vec4& a, const vec4& b);
	friend vec4 operator -(const vec4& a, const vec4& b);
	friend vec4 operator *(const vec4& a, const double d );
	friend vec4 operator *(const double d, const vec4& a);
	friend vec4 operator *(const mat4& a, const vec4& v);
	friend vec4 operator *(const vec4& v, mat4& a);
	friend double operator *(const vec4& a, const vec4& b);
	friend vec4 operator /(const vec4& a, const double d);
	friend bool operator ==(const vec4& a, const vec4& b);
	friend bool operator !=(const vec4& a, const vec4& b);
	friend ostream& operator <<(ostream& s, const vec4& v);
	friend istream& operator >>(istream& s, vec4& v);
	friend void swap(vec4& a, vec4& b);	
	friend vec4 minimum(const vec4& a, const vec4& b);
	friend vec4 maximum(const vec4& a, const vec4& b);
	friend vec4 prod(const vec4& a, const vec4& b);

	friend class mat3;
	friend class mat4;

	friend mat4 operator * (mat4& a, mat4& b);
#endif
public:
	double n[4];
};

class mat3
{
public:
	mat3()
		{ v[0] = vec3(); v[1] = vec3(); v[2] = vec3();
		  v[0][0] = 1.0; v[1][1] = 1.0; v[2][2] = 1.0; }
	mat3( const vec3& v0, const vec3& v1, const vec3& v2 )
		{ v[0] = v0; v[1] = v1; v[2] = v2; }
//	mat3( const double d )
//		{ v[0] = vec3(); v[1] = vec3(); v[2] = vec3();
//		  v[0][0] = d; v[1][1] = d; v[2][2] = d; }
	mat3( const mat3& m )
		{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; }

	mat3& operator =( const mat3& m )
		{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; return *this; }
	mat3& operator +=( const mat3& m )
		{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; return *this; }
	mat3& operator -=( const mat3& m )
		{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; return *this; }
	mat3& operator *=( const double d )
		{ v[0] *= d; v[1] *= d; v[2] *= d; return *this; }
	mat3& operator /=( const double d )
		{ v[0] /= d; v[1] /= d; v[2] /= d; return *this; }

	vec3& operator []( int i )
		{ return v[i]; }
	const vec3& operator []( int i ) const
		{ return v[i]; }
	
	vec3 column( int i ) const
		{ return vec3( v[0][i], v[1][i], v[2][i] ); }

	// special functions

	mat3 transpose() const
	{
		return mat3( column( 0 ), column( 1 ), column( 2 ) );
	}

	mat3 inverse() const;
#if 0
	friend mat3 operator -( const mat3& a );
	friend mat3 operator +( const mat3& a, const mat3& b );
	friend mat3 operator -( const mat3& a, const mat3& b);
	friend mat3 operator *( const mat3& a, const mat3& b );
	friend mat3 operator *( const mat3& a, const double d );
	friend mat3 operator *( const double d, const mat3& a );
	friend mat3 operator /( const mat3& a, const double d );
	friend bool operator ==( const mat3& a, const mat3& b );
	friend bool operator !=( const mat3& a, const mat3& b );
	friend ostream& operator <<( ostream& s, const mat3& m );
	friend istream& operator >>( istream& s, mat3& m );
	friend void swap(mat3& a, mat3& b);

	// necessary friend declarations

	friend vec3 operator * (const mat3& a, const vec3& v);	    // linear transform
#endif
public:
	vec3 v[3];
};

class mat4
{
public:
	mat4()
		{ v[0]=vec4(); v[1]=vec4(); v[2]=vec4(); v[3]=vec4();
		  v[0][0]=1.0; v[1][1]=1.0; v[2][2]=1.0; v[3][3]=1.0; }
	mat4( const vec4& v0, const vec4& v1, const vec4& v2, const vec4& v3 )
		{ v[0] = v0; v[1] = v1; v[2] = v2; v[3] = v3; }
//	mat4( const double d )
//		{ v[0]=vec4(); v[1]=vec4(); v[2]=vec4(); v[3]=vec4();
//		  v[0][0]=d; v[1][1]=d; v[2][2]=d; v[3][3]=d; }
	mat4( const mat4& m )
		{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3]; }

	mat4& operator =( const mat4& m )
		{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3];
		  return *this; }
	mat4& operator +=( const mat4& m )
		{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; v[3] += m.v[3];
		  return *this; }
	mat4& operator -=( const mat4& m )
		{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; v[3] -= m.v[3];
		  return *this; }
	mat4& operator *=( const double d )
		{ v[0] *= d; v[1] *= d; v[2] *= d; v[3] *= d; return *this; }
	mat4& operator /=( const double d )
		{ v[0] /= d; v[1] /= d; v[2] /= d; v[3] /= d; return *this; }

	vec4& operator []( int i )
		{ return v[i]; }
	const vec4& operator []( int i ) const
		{ return v[i]; }
	vec4 column( int i ) const
		{ return vec4( v[0][i], v[1][i], v[2][i], v[3][i] ); }

	mat4 transpose() const
		{ return mat4( column( 0 ), column( 1 ), column( 2 ), column( 3 ) ); }
	mat4 inverse() const;
	mat3 upper33() const
		{ return mat3( vec3( v[0] ), vec3( v[1] ), vec3( v[2] ) ); }
#if 0
	friend mat4 operator -( const mat4& a );
	friend mat4 operator +( const mat4& a, const mat4& b );
	friend mat4 operator -( const mat4& a, const mat4& b );
	friend mat4 operator *( const mat4& a, const mat4& b );
	friend mat4 operator *( const mat4& a, const double d );
	friend mat4 operator *( const double d, const mat4& a );
	friend mat4 operator /( const mat4& a, const double d );
	friend bool operator ==( const mat4& a, const mat4& b );
	friend bool operator !=( const mat4& a, const mat4& b );
	friend ostream& operator <<( ostream& s, const mat4& m );
	friend istream& operator >>( istream& s, mat4& m );
	friend void swap( mat4& a, mat4& b );

	friend vec4 operator * (const mat4& a, const vec4& v);	    // linear transform
	friend vec3 operator * (const mat4& a, const vec3& v);	    // linear transform
#endif
	static mat4 identity()
	{ return mat4(
		vec4( 1.0, 0.0, 0.0, 0.0 ),
		vec4( 0.0, 1.0, 0.0, 0.0 ),
		vec4( 0.0, 0.0, 1.0, 0.0 ),
		vec4( 0.0, 0.0, 0.0, 1.0 )); }

	static mat4 translate( const vec3& v )
	{ return mat4(
		vec4( 1.0, 0.0, 0.0, v[0] ),
		vec4( 0.0, 1.0, 0.0, v[1] ),
		vec4( 0.0, 0.0, 1.0, v[2] ),
		vec4( 0.0, 0.0, 0.0, 1.0 )); }

	static mat4 rotate( const vec3& axis, const double angle ) { 
		double c = cos( angle );
		double s = sin( angle );
		double t = 1.0 - c;

		vec3 a = axis.normalize();
		return mat4(
			vec4(t*a[0]*a[0]+c, t*a[0]*a[1]-s*a[2], t*a[0]*a[2]+s*a[1], 0.0),
			vec4(t*a[0]*a[1]+s*a[2], t*a[1]*a[1]+c, t*a[1]*a[2]-s*a[0], 0.0),
			vec4(t*a[0]*a[2]-s*a[1], t*a[1]*a[2]+s*a[0], t*a[2]*a[2]+c, 0.0),
			vec4(0.0, 0.0, 0.0, 1.0) );
	}

	static mat4 scale( const vec3& t )
	{ return mat4(
		vec4( t[0], 0.0, 0.0, 0.0 ),
		vec4( 0.0, t[1], 0.0, 0.0 ),
		vec4( 0.0, 0.0, t[2], 0.0 ),
		vec4( 0.0, 0.0, 0.0, 1.0 )); }

	static mat4 perspective3D( const double d )
	{ return mat4(
		vec4( 1.0, 0.0, 0.0, 0.0 ),
		vec4( 0.0, 1.0, 0.0, 0.0 ),
		vec4( 0.0, 0.0, 1.0, 0.0 ),
		vec4( 0.0, 0.0, 1.0/d, 0.0 )); }

public:
	vec4 v[4];
};

/****************************************************************
*								*
*	       2D functions and 3D functions			*
*								*
****************************************************************/

mat3 identity2D();					    // identity 2D
mat4 identity3D();					    // identity 3D
mat4 translation3D(vec3& v);				    // translation 3D
mat4 rotation3D(vec3& Axis, const double angleDeg);	    // rotation 3D
mat4 scaling3D(vec3& scaleVector);			    // scaling 3D
mat4 perspective3D(const double d);			    // perspective 3D

// And now, many inline functions are defined.

inline double operator *( const vec3& a, const vec4& b )
{
	return a[0] * b[0] + a[1] * b[1] + a[2] * b[2] + b[3];
}

inline double operator *( const vec4& b, const vec3& a )
{
	return a[0] * b[0] + a[1] * b[1] + a[2] * b[2] + b[3];
}

inline vec3 operator -(const vec3& v)
{
	return vec3( -v.n[0], -v.n[1], -v.n[2] );
}

inline vec3 operator +(const vec3& a, const vec3& b)
{
	return vec3( a.n[0] + b.n[0], a.n[1] + b.n[1], a.n[2] + b.n[2] );
}

inline vec3 operator -(const vec3& a, const vec3& b)
{
	return vec3( a.n[0] - b.n[0], a.n[1] - b.n[1], a.n[2] - b.n[2] );
}

inline vec3 operator *(const vec3& a, const double d )
{
	return vec3( a.n[0] * d, a.n[1] * d, a.n[2] * d );
}

inline vec3 operator *(const double d, const vec3& a)
{
	return a * d;
}

inline vec3 operator *(const mat4& a, const vec3& v)
{
	return vec3( a[0] * v, a[1] * v, a[2] * v ); 
}

inline vec3 operator *(const vec3& v, mat4& a)
{
	return a.transpose() * v;
}

inline double operator *(const vec3& a, const vec3& b)
{
	return a.n[0]*b.n[0] + a.n[1]*b.n[1] + a.n[2]*b.n[2];
}

inline vec3 operator *( const mat3& a, const vec3& b )
{
	return vec3( a[0]*b, a[1]*b, a[2]*b );
}

inline vec3 operator *( const vec3& a, const mat3& b )
{
	return vec3( b.column(0)*a, b.column(1)*a, b.column(2)*a );
}

inline vec3 operator /(const vec3& a, const double d)
{
	return vec3( a.n[0] / d, a.n[1] / d, a.n[2] / d );
}

inline vec3 operator ^(const vec3& a, const vec3& b)
{
	return vec3(
		a.n[1]*b.n[2] - a.n[2]*b.n[1],
		a.n[2]*b.n[0] - a.n[0]*b.n[2],
		a.n[0]*b.n[1] - a.n[1]*b.n[0] );
}

inline bool operator ==(const vec3& a, const vec3& b)
{
	return a.n[0]==b.n[0] && a.n[1] == b.n[1] && a.n[2] == b.n[2];
}

inline bool operator !=(const vec3& a, const vec3& b)
{
	return !( a == b );
}

inline ostream& operator <<( ostream& os, const vec3& v )
{
	return os << v.n[0] << " " << v.n[1] << " " << v.n[2];
}

inline istream& operator >>( istream& is, vec3& v )
{
	return is >> v.n[0] >> v.n[1] >> v.n[2];
}

inline void swap( vec3& a, vec3& b )
{
	vec3 t( a );
	a = b;
	b = t;
}

inline vec3 minimum( const vec3& a, const vec3& b )
{
	return vec3( minimum(a.n[0],b.n[0]), minimum(a.n[1],b.n[1]), minimum(a.n[2],b.n[2]) );
}

inline vec3 maximum(const vec3& a, const vec3& b)
{
	return vec3( maximum(a.n[0],b.n[0]), maximum(a.n[1],b.n[1]), maximum(a.n[2],b.n[2]) );
}

inline vec3 prod(const vec3& a, const vec3& b )
{
	return vec3( a.n[0]*b.n[0], a.n[1]*b.n[1], a.n[2]*b.n[2] );
}

inline vec4 operator -( const vec4& v )
{
	return vec4( -v.n[0], -v.n[1], -v.n[2], -v.n[3] );
}

inline vec4 operator +( const vec4& a, const vec4& b )
{
	return vec4( a.n[0] + b.n[0], a.n[1] + b.n[1], a.n[2] + b.n[2],
		a.n[3] + b.n[3] );
}

inline vec4 operator -(const vec4& a, const vec4& b)
{
	return vec4( a.n[0] - b.n[0], a.n[1] - b.n[1], a.n[2] - b.n[2],
		a.n[3] - b.n[3] );
}

inline vec4 operator *(const vec4& a, const double d )
{
	return vec4( a.n[0] * d, a.n[1] * d, a.n[2] * d, a.n[3] * d );
}

inline vec4 operator *(const double d, const vec4& a)
{
	return a * d;
}

inline double operator *(const vec4& a, const vec4& b)
{
	return a.n[0]*b.n[0] + a.n[1]*b.n[1] + a.n[2]*b.n[2] + a.n[3]*b.n[3];
}

inline vec4 operator *(const mat4& a, const vec4& v)
{
	return vec4( a[0] * v, a[1] * v, a[2] * v, a[3] * v );
}

inline vec4 operator *( const vec4& v, mat4& a )
{
	return a.transpose() * v;
}

inline vec4 operator /(const vec4& a, const double d)
{
	return vec4( a.n[0] / d, a.n[1] / d, a.n[2] / d, a.n[3] / d );
}

inline bool operator ==(const vec4& a, const vec4& b)
{
	return a.n[0] == b.n[0] && a.n[1] == b.n[1] && a.n[2] == b.n[2] 
	    && a.n[3] == b.n[3];
}

inline bool operator !=(const vec4& a, const vec4& b)
{
	return !( a == b );
}

inline ostream& operator <<( ostream& os, const vec4& v )
{
	return os << v.n[0] << " " << v.n[1] << " " << v.n[2] << " " << v.n[3];
}

inline istream& operator >>( istream& is, vec4& v )
{
	return is >> v.n[0] >> v.n[1] >> v.n[2] >> v.n[3];
}

inline void swap( vec4& a, vec4& b )
{
	vec4 t( a );
	a = b;
	b = t;
}

inline vec4 minimum( const vec4& a, const vec4& b )
{
	return vec4( minimum(a.n[0],b.n[0]), minimum(a.n[1],b.n[1]), minimum(a.n[2],b.n[2]),
	             minimum(a.n[3],b.n[3]) );
}

inline vec4 maximum(const vec4& a, const vec4& b)
{
	return vec4( maximum(a.n[0],b.n[0]), maximum(a.n[1],b.n[1]), maximum(a.n[2],b.n[2]),
	             maximum(a.n[3],b.n[3]) );
}

inline vec4 prod(const vec4& a, const vec4& b )
{
	return vec4( a.n[0]*b.n[0], a.n[1]*b.n[1], a.n[2]*b.n[2], a.n[3]*b.n[3] );
}

inline mat3 operator -( const mat3& a )
{
	return mat3( -a.v[0], -a.v[1], -a.v[2] );
}

inline mat3 operator +( const mat3& a, const mat3& b )
{
	return mat3( a.v[0]+b.v[0], a.v[1]+b.v[1], a.v[2]+b.v[2] );
}

inline mat3 operator -( const mat3& a, const mat3& b)
{
	return mat3( a.v[0]-b.v[0], a.v[1]-b.v[1], a.v[2]-b.v[2] );
}

inline mat3 operator *( const mat3& a, const mat3& b )
{
	vec3 c0 = b.column( 0 );
	vec3 c1 = b.column( 1 );
	vec3 c2 = b.column( 2 );

	return mat3( 
		vec3( a.v[0]*c0, a.v[0]*c1, a.v[0]*c2 ),
		vec3( a.v[1]*c0, a.v[1]*c1, a.v[1]*c2 ),
		vec3( a.v[2]*c0, a.v[2]*c1, a.v[2]*c2 ) );
}

inline mat3 operator *( const mat3& a, const double d )
{
	return mat3( a.v[0]*d, a.v[1]*d, a.v[2]*d );
}

inline mat3 operator *( const double d, const mat3& a )
{
	return mat3( d*a.v[0], d*a.v[1], d*a.v[2] );
}

inline mat3 operator /( const mat3& a, const double d )
{
	return mat3( a.v[0]/d, a.v[1]/d, a.v[2]/d );
}

inline bool operator ==( const mat3& a, const mat3& b )
{
	return a.v[0]==b.v[0] && a.v[1]==b.v[1] && a.v[2]==b.v[2];
}

inline bool operator !=( const mat3& a, const mat3& b )
{
	return !( a == b );
}

inline ostream& operator <<( ostream& os, const mat3& m )
{
	return os << m.v[0] << " " << m.v[1] << " " << m.v[2];
}

inline istream& operator >>( istream& is, mat3& m )
{
	return is >> m.v[0] >> m.v[1] >> m.v[2];
}

inline void swap(mat3& a, mat3& b)
{
	swap( a.v[0], b.v[0] );
	swap( a.v[1], b.v[1] );
	swap( a.v[2], b.v[2] );
}

inline mat4 operator -( const mat4& a )
{
	return mat4( -a.v[0], -a.v[1], -a.v[2], -a.v[3] );
}

inline mat4 operator +( const mat4& a, const mat4& b )
{
	return mat4( a.v[0]+b.v[0], a.v[1]+b.v[1], a.v[2]+b.v[2], a.v[3]+b.v[3] );
}

inline mat4 operator -( const mat4& a, const mat4& b )
{
	return mat4( a.v[0]-b.v[0], a.v[1]-b.v[1], a.v[2]-b.v[2], a.v[3]-b.v[3] );
}

inline mat4 operator *( const mat4& a, const mat4& b )
{
	vec4 c0 = b.column( 0 );
	vec4 c1 = b.column( 1 );
	vec4 c2 = b.column( 2 );
	vec4 c3 = b.column( 3 );

	return mat4( 
		vec4( a.v[0]*c0, a.v[0]*c1, a.v[0]*c2, a.v[0]*c3 ),
		vec4( a.v[1]*c0, a.v[1]*c1, a.v[1]*c2, a.v[1]*c3 ),
		vec4( a.v[2]*c0, a.v[2]*c1, a.v[2]*c2, a.v[2]*c3 ),
		vec4( a.v[3]*c0, a.v[3]*c1, a.v[3]*c2, a.v[3]*c3 ) );
}

inline mat4 operator *( const mat4& a, const double d )
{
	return mat4( a.v[0]*d, a.v[1]*d, a.v[2]*d, a.v[3]*d );
}

inline mat4 operator *( const double d, const mat4& a )
{
	return mat4( d*a.v[0], d*a.v[1], d*a.v[2], d*a.v[3] );
}

inline mat4 operator /( const mat4& a, const double d )
{
	return mat4( a.v[0]/d, a.v[1]/d, a.v[2]/d, a.v[3]/d );
}

inline bool operator ==( const mat4& a, const mat4& b )
{
	return a.v[0]==b.v[0] && a.v[1]==b.v[1] && a.v[2]==b.v[2] && a.v[3]==b.v[3];
}

inline bool operator !=( const mat4& a, const mat4& b )
{
	return !( a == b );
}

inline ostream& operator <<( ostream& os, const mat4& m )
{
	return os << m.v[0] << " " << m.v[1] << " " << m.v[2] << " " << m.v[3];
}

inline istream& operator >>( istream& is, mat4& m )
{
	return is >> m.v[0] >> m.v[1] >> m.v[2] >> m.v[3];
}

inline void swap( mat4& a, mat4& b )
{
	swap( a.v[0], b.v[0] );
	swap( a.v[1], b.v[1] );
	swap( a.v[2], b.v[2] );
	swap( a.v[3], b.v[3] );
}

inline vec3::vec3( const vec4& v ) 
{ 
	n[0] = v[0]; 
	n[1] = v[1]; 
	n[2] = v[2]; 
}

inline vec3 clamp( const vec3& other )
{
	return maximum( vec3(), minimum( other, vec3( 1.0, 1.0, 1.0 ) ) );
}

#endif // __ALGEBRA_H__
