// 
// scene.cpp
// 
// GEOMETRY.  The main contents of this file are all the ray-object
// intersection tests for different geometric primitives.   This is 
// an annoyingly difficult task.  There are plenty of opportunities 
// to make silly mistakes, and this code probably still contains many.
// Furthermore, there's a big desire to make this code fast, since
// these intersection tests are called an enormous number of times
// in any trace.  But making them faster also makes them more complicated
// and increases the risk of bugs!
//
// In short, that's why we give this code to you.  The experience of
// writing these methods yourself would be character building at best,
// but wouldn't teach all that much about graphics.

#include <cmath>

#include "scene.h"
#include "light.h"

bool Geometry::intersect( const ray& r, isect& i ) const
{
	return false;
}

bool Geometry::intersect( const ray& r ) const
{
	isect i;
	return intersect( r, i );
}

bool Sphere::intersect( const ray& r, isect& i ) const
{
	vec3 v = -r.getPosition();
	double b = v * r.getDirection();
	double disc = b*b - v*v + 1;

	if( disc < 0.0 ) {
		return false;
	}

	disc = sqrt( disc );
	double t2 = b + disc;

	if( t2 <= RayGlobals::ray_epsilon ) {
		return false;
	}

	i.obj = this;

	double t1 = b - disc;

	if( t1 > RayGlobals::ray_epsilon ) {
		i.t = t1;
		i.N = r.at( t1 ).normalize();
	} else {
		i.t = t2;
		i.N = r.at( t2 ).normalize();
	}

	return true;
}

bool Sphere::intersect( const ray& r ) const
{
	vec3 v = -r.getPosition();
	double b = v * r.getDirection();
	double disc = b*b - v*v + 1;

	if( disc < 0.0 ) {
		return false;
	}

	return (b + sqrt( disc )) > RayGlobals::ray_epsilon;
}

bool Box::intersect( const ray& r, isect& i ) const
{
	double tnear = -1.0E9;
	double tfar = 1.0E9;

/*
	bool have_near = false;
	bool have_far = false;
*/

	int near_slab = 0;
	int far_slab = 0;

	const vec3& P = r.getPosition();
	const vec3& D = r.getDirection();

	for( int idx = 0; idx < 3; ++idx ) {
		if( D[idx] == 0.0 ) {
			if( P[idx] < -0.5 || P[idx] > 0.5 ) {
				return false;
			}
		} else {
			double t1c = -(0.5 + P[idx]) / D[idx];
			double t2c = (0.5 - P[idx]) / D[idx];

			if( t1c > t2c ) {
				swap( t1c, t2c );
			}

/*
			if( have_far ) {
*/
				if( t2c < tfar ) {
					tfar = t2c;
					far_slab = idx;
				} 
/*
			} else {
				tfar = t2c;
				far_slab = idx;
				have_far = true;
			}
*/

/*
			if( have_near ) {
*/
				if( t1c > tnear ) {
					tnear = t1c;
					near_slab = idx;
				}
/*
			} else {
				tnear = t1c;
				near_slab = idx;
				have_near = true;
			}
*/

			if( tnear > tfar ) {
				return false;
			}

			if( tfar < RayGlobals::ray_epsilon ) {
				return false;
			}
		}
	}

/*
	if( !have_near ) {
		return false;
	}
*/

	if( tnear < RayGlobals::ray_epsilon ) {
		tnear = tfar;
		near_slab = far_slab;
	}

	i.obj = this;
	i.t = tnear;
	i.material = NULL;

	vec3 N( 0.0, 0.0, 0.0 );
	N[near_slab] = 1.0;

	if( P[near_slab] < -0.5 ) {
		i.N = -N;
	} else if( P[near_slab] > 0.5 ) {
		i.N = N;
	} else {
		if( D[near_slab] < 0.0 ) {
			i.N = -N;
		} else {
			i.N = N;
		}
	}

	return true;
}

bool Square::intersect( const ray& r, isect& i ) const
{
	vec3 p = r.getPosition();
	vec3 d = r.getDirection();

	if( d[2] == 0.0 ) {
		return false;
	}

	double t = -p[2]/d[2];

	if( t <= RayGlobals::ray_epsilon ) {
		return false;
	}

	vec3 P = r.at( t );

	if( P[0] < -0.5 || P[0] > 0.5 ) {	
		return false;
	}

	if( P[1] < -0.5 || P[1] > 0.5 ) {	
		return false;
	}

	i.obj = this;
	i.t = t;
	i.N = vec3( 0.0, 0.0, 1.0 );

	return true;
}

bool Cylinder::intersect( const ray& r, isect& i ) const
{
	if( intersectCaps( r, i ) ) {
		isect ii;
		if( intersectBody( r, ii ) ) {
			if( ii.t < i.t ) {
				i = ii;
			}
		}
		return true;
	} else {
		return intersectBody( r, i );
	}
}

bool Cylinder::intersectBody( const ray& r, isect& i ) const
{
	double x0 = r.getPosition()[0];
	double y0 = r.getPosition()[1];
	double x1 = r.getDirection()[0];
	double y1 = r.getDirection()[1];

	double a = 2.0*(x1*x1+y1*y1);
	double b = -2.0*(x0*x1 + y0*y1);
	double c = x0*x0 + y0*y0 - 1.0;

	if( 0.0 == a ) {
		// This implies that x1 = 0.0 and y1 = 0.0, which further
		// implies that the ray is aligned with the body of the cylinder,
		// so no intersection.
		return false;
	}

	double disc = b*b - 2.0*a*c;

	if( disc < 0.0 ) {
		return false;
	}
	
	disc = sqrt( disc );

	double t2 = (b + disc) / a;

	if( t2 <= RayGlobals::ray_epsilon ) {
		return false;
	}

	double t1 = (b - disc) / a;

	if( t1 > RayGlobals::ray_epsilon ) {
		// Two intersections.
		vec3 P = r.at( t1 );
		double z = P[2];
		if( z >= 0.0 && z <= 1.0 ) {
			// It's okay.
			i.obj = this;
			i.t = t1;
			i.N = vec3( P[0], P[1], 0.0 ).normalize();
			return true;
		}
	}

	vec3 P = r.at( t2 );
	double z = P[2];
	if( z >= 0.0 && z <= 1.0 ) {
		i.obj = this;
		i.t = t2;
		i.N = vec3( P[0], P[1], 0.0 ).normalize();
		return true;
	}

	return false;
}

bool Cylinder::intersectCaps( const ray& r, isect& i ) const
{
	if( !capped ) {
		return false;
	}

	double pz = r.getPosition()[2];
	double dz = r.getDirection()[2];

	if( 0.0 == dz ) {
		return false;
	}

	double t1;
	double t2;

	if( dz > 0.0 ) {
		t1 = (-pz)/dz;
		t2 = (1.0-pz)/dz;
	} else {
		t1 = (1.0-pz)/dz;
		t2 = (-pz)/dz;
	}

	if( t2 < RayGlobals::ray_epsilon ) {
		return false;
	}

	if( t1 >= RayGlobals::ray_epsilon ) {
		vec3 p( r.at( t1 ) );
		if( (p[0]*p[0] + p[1]*p[1]) <= 1.0 ) {
			i.obj = this;
			i.t = t1;
			if( dz > 0.0 ) {
				// Intersection with cap at z = 0.
				i.N = vec3( 0.0, 0.0, -1.0 );
			} else {
				i.N = vec3( 0.0, 0.0, 1.0 );
			}
			return true;
		}
	}

	vec3 p( r.at( t2 ) );
	if( (p[0]*p[0] + p[1]*p[1]) <= 1.0 ) {
		i.obj = this;
		i.t = t2;
		if( dz > 0.0 ) {
			// Intersection with interior of cap at z = 1.
			i.N = vec3( 0.0, 0.0, -1.0 );
		} else {
			i.N = vec3( 0.0, 0.0, 1.0 );
		}
		return true;
	}

	return false;
}

bool Cone::intersect( const ray& r, isect& i ) const
{
	if( intersectCaps( r, i ) ) {
		isect ii;
		if( intersectBody( r, ii ) ) {
			if( ii.t < i.t ) {
				i = ii;
			}
		}
		return true;
	} else {
		return intersectBody( r, i );
	}
}

bool Cone::intersectBody( const ray& r, isect& i ) const
{
	vec3 d = r.getDirection();
	vec3 p = r.getPosition();

	double a = (d[0]*d[0]) + (d[1]*d[1]) - (C*d[2]*d[2]);
	double b = 2.0 * (d[0]*p[0] + d[1]*p[1] - C*d[2]*p[2]) - B*d[2];
	double c = (p[0]*p[0]) + (p[1]*p[1]) - A - (B*p[2]) - (C*p[2]*p[2]);

	double disc = b*b - 4.0*a*c;

	if( disc <= 0.0 ) {
		return false;
	}

	disc = sqrt( disc );

	double t1 = (-b - disc) / (2.0 * a);
	double t2 = (-b + disc) / (2.0 * a);

	if( t2 < RayGlobals::ray_epsilon ) {
		return false;
	}

	if( t1 > RayGlobals::ray_epsilon ) {
		// Two intersections.
		vec3 P = r.at( t1 );
		double z = P[2];
		if( z >= 0.0 && z <= height ) {
			// It's okay.
			i.obj = this;
			i.t = t1;
			i.N = vec3( P[0], P[1], 
				P.length() * (b_radius - t_radius) / height ).normalize();
			return true;
		}
	}

	vec3 P = r.at( t2 );
	double z = P[2];
	if( z >= 0.0 && z <= height ) {
		i.obj = this;
		i.t = t2;
		i.N = vec3( P[0], P[1], 
			P.length() * (b_radius - t_radius) / height ).normalize();
		return true;
	}

	return false;
}

bool Cone::intersectCaps( const ray& r, isect& i ) const
{
	if( !capped ) {
		return false;
	}

	double pz = r.getPosition()[2];
	double dz = r.getDirection()[2];

	if( 0.0 == dz ) {
		return false;
	}

	double t1;
	double t2;
	double r1;
	double r2;

	if( dz > 0.0 ) {
		t1 = (-pz)/dz;
		t2 = (height-pz)/dz;
		r1 = b_radius;
		r2 = t_radius;
	} else {
		t1 = (height-pz)/dz;
		t2 = (-pz)/dz;
		r1 = t_radius;
		r2 = b_radius;
	}

	if( t2 < RayGlobals::ray_epsilon ) {
		return false;
	}

	if( t1 >= RayGlobals::ray_epsilon ) {
		vec3 p( r.at( t1 ) );
		if( (p[0]*p[0] + p[1]*p[1]) <= r1 * r1 ) {
			i.obj = this;
			i.t = t1;
			if( dz > 0.0 ) {
				// Intersection with cap at z = 0.
				i.N = vec3( 0.0, 0.0, -1.0 );
			} else {
				i.N = vec3( 0.0, 0.0, 1.0 );
			}
			return true;
		}
	}

	vec3 p( r.at( t2 ) );
	if( (p[0]*p[0] + p[1]*p[1]) <= r2 * r2 ) {
		i.obj = this;
		i.t = t2;
		if( dz > 0.0 ) {
			// Intersection with interior of cap at z = 1.
			i.N = vec3( 0.0, 0.0, -1.0 );
		} else {
			i.N = vec3( 0.0, 0.0, 1.0 );
		}
		return true;
	}

	return false;
}

bool TransformedGeometry::intersect( const ray& r, isect& i ) const
{
	vec3 pp = inverse * r.getPosition();
	vec3 dp = inverse * (r.getPosition() + r.getDirection()) - pp;
	double len = dp.length();
	dp /= len;

	if( sub->intersect( ray( pp, dp ), i ) ) {
		i.obj = this;
		i.N = (normi * i.N).normalize();
		i.t /= len;

		return true;
	}

	return false;
}

bool TransformedGeometry::intersect( const ray& r ) const
{
	vec3 pp = inverse * r.getPosition();
	vec3 dp = (inverse * (r.getPosition() + r.getDirection()) - pp).normalize();
	return sub->intersect( ray( pp, dp ) );
}

SceneObject *TransformedGeometry::compact()
{
	SceneObject *sg = sub->compact();

	if( sg->isTransformed() ) {
		TransformedGeometry *tg = dynamic_cast<TransformedGeometry*>( sg );

		tg->xform = xform * tg->xform;
		tg->inverse = tg->xform.inverse();
		tg->normi = tg->xform.upper33().inverse().transpose();

		sub = NULL;
		delete this;
		return sg;
	} else {
		if( sg != sub ) {
			delete sub;
			sub = sg;
		}

		return this;
	}
}

Scene::~Scene()
{
	for( giter g = objects.begin(); g != objects.end(); ++g ) {
		delete (*g);
	}

	for( liter l = lights.begin(); l != lights.end(); ++l ) {
		delete (*l);
	}
}

// Get any intersection with an object.  Return information about the 
// intersection through the reference parameter.
bool Scene::intersect( const ray& r, isect& i ) const
{
	typedef list<Geometry*>::const_iterator iter;

	isect cur;
	bool have_one = false;

	for( iter j = objects.begin(); j != objects.end(); ++j ) {
		if( (*j)->intersect( r, cur ) ) {
			if( !have_one || (cur.t < i.t) ) {
				i = cur;
				have_one = true;
			}
		}
	}

	if( have_one ) {
		// Now, fill in the indices of refraction for the object that
		// got the intersection.  This calculation is horribly inadequate
		// for handling refraction in the case of intersecting objects.  Yech!

		if( r.getDirection() * i.N < 0.0 ) {
			// greater than 90 degrees separation between normal and incident
			// ray.  We must be entering the surface.
			i.setIndices( 1.0, i.getMaterial().index );
		} else {
			// ray lined up with normal.  We must be leaving the surface.
			i.setIndices( i.getMaterial().index, 1.0 );

			// Also, reverse the normal to point back into the surface.
			i.N = -i.N;
		}

		return true;
	} else {
		return false;
	}
}

// Get any intersection with an object.  Return whether or not an 
// intersection occured.
bool Scene::intersect( const ray& r ) const
{
	typedef list<Geometry*>::const_iterator iter;
	for( iter j = objects.begin(); j != objects.end(); ++j ) {
		Geometry *geo = (*j);
		if( geo->intersect( r ) ) {
			return true;
		}
	}
	return false;
}

// Get any intersection with an object.  Return whether or not an 
// intersection occured within the distance tmax from the origin of the ray.
bool Scene::intersect( const ray& r, double tmax ) const
{
	typedef list<Geometry*>::const_iterator iter;
	isect i;
	for( iter j = objects.begin(); j != objects.end(); ++j ) {
		if( (*j)->intersect( r, i ) ) {
			if( i.t < tmax ) {
				// cout << "\tgot isect at t = " << i.t << endl;
				return true;
			}
		}
	}
	return false;
}
