//
// scene.h
//
// The Scene class and the geometric types that it can contain.
//

#ifndef __SCENE_H__
#define __SCENE_H__

#include <list>
#include <algorithm>

using namespace std;

#include "general.h"
#include "ray.h"
#include "material.h"
#include "camera.h"

class Light;
class Scene;

class SceneElement
{
public:
	virtual ~SceneElement() {}

	Scene *getScene() const { return scene; }

protected:
	SceneElement( Scene *s )
		: scene( s ) {}

private:
	Scene *scene;
};

// A Geometry object is anything that has extent in three dimensions.
// It may not be an actual visible scene object.  For example, hierarchical
// spatial subdivision could be expressed in terms of Geometry instances.
class Geometry
	: public SceneElement
{
public:
	virtual bool intersect( const ray& r, isect& i ) const;
	virtual bool intersect( const ray& r ) const;

	Geometry( Scene *scene ) 
		: SceneElement( scene ) {}
};

// A SceneObject is a real actual thing that we want to model in the 
// world.  It has extent (its Geometry heritage) and surface properties
// (its material binding).  The decision of how to store that material
// is left up to the subclass.
class SceneObject
	: public Geometry
{
public:
	virtual const Material& getMaterial() const = 0;
	virtual void setMaterial( Material *m ) = 0;

	virtual bool isTransformed() const { return false; }
	virtual SceneObject *compact() { return this; }

protected:
	SceneObject( Scene *scene )
		: Geometry( scene ) {}
};

// A simple extension of SceneObject that adds an instance of Material
// for simple material bindings.
class MaterialSceneObject
	: public SceneObject
{
public:
	virtual ~MaterialSceneObject() { if( material ) delete material; }

	virtual const Material& getMaterial() const { return *material; }
	virtual void setMaterial( Material *m )	{ material = m; }

protected:
	MaterialSceneObject( Scene *scene, Material *mat ) 
		: SceneObject( scene ), material( mat ) {}
	MaterialSceneObject( Scene *scene ) 
		: SceneObject( scene ), material( new Material ) {}

	Material *material;
};

class Sphere
	: public MaterialSceneObject
{
public:
	Sphere( Scene *scene, Material *mat )
		: MaterialSceneObject( scene, mat ) {}
	Sphere( Scene *scene )
		: MaterialSceneObject( scene ) {}

	virtual bool intersect( const ray& r, isect& i ) const;
	virtual bool intersect( const ray& r ) const;
};

class Box
	: public MaterialSceneObject
{
public:
	Box( Scene *scene, Material *mat )
		: MaterialSceneObject( scene, mat ) {}
	Box( Scene *scene )
		: MaterialSceneObject( scene ) {}

	virtual bool intersect( const ray& r, isect& i ) const;
};

class Square
	: public MaterialSceneObject
{
public:
	Square( Scene *scene, Material *mat )
		: MaterialSceneObject( scene, mat ) {}
	Square( Scene *scene )
		: MaterialSceneObject( scene ) {}

	virtual bool intersect( const ray& r, isect& i ) const;
};

class Cylinder
	: public MaterialSceneObject
{
public:
	Cylinder( Scene *scene, Material *mat )
		: MaterialSceneObject( scene, mat ), capped( true ) {}
	Cylinder( Scene *scene )
		: MaterialSceneObject( scene ), capped( true ) {}
	
	virtual bool intersect( const ray& r, isect& i ) const;

	bool intersectBody( const ray& r, isect& i ) const;
	bool intersectCaps( const ray& r, isect& i ) const;

protected:
	bool capped;
};

class Cone
	: public MaterialSceneObject
{
public:
	Cone( Scene *scene, Material *mat, 
			double h = 1.0, double br = 1.0, double tr = 0.0, 
			bool cap = false )
		: MaterialSceneObject( scene, mat )
		, capped( cap ) 
		, height( h )
		, b_radius( br )
		, t_radius( tr )
	{ computeABC(); }

	Cone( Scene *scene, 
			double h = 1.0, double br = 1.0, double tr = 0.0, 
			bool cap = false )
		: MaterialSceneObject( scene )
		, capped( cap )
		, height( h )
		, b_radius( br )
		, t_radius( tr )
	{ computeABC(); }
	
	virtual bool intersect( const ray& r, isect& i ) const;

	bool intersectBody( const ray& r, isect& i ) const;
	bool intersectCaps( const ray& r, isect& i ) const;

protected:
	void computeABC()
	{
		A = b_radius * b_radius;
		B = 2.0 * b_radius * (t_radius - b_radius) / height;
		C = (t_radius - b_radius) / height;
		C = C * C;
	}

	bool capped;
	double height;
	double b_radius;
	double t_radius;

	double A;
	double B;
	double C;
};

// A transformed version of some lower-level geometry.  Wraps the intersection
// tests of its child with transforms.
class TransformedGeometry
	: public SceneObject
{
public:
	TransformedGeometry( SceneObject *obj, const mat4& mat )
		: SceneObject( obj->getScene() ), sub( obj ), xform( mat )  
	{
		inverse = xform.inverse();
		normi = xform.upper33().inverse().transpose();
	}
	virtual ~TransformedGeometry() { if( sub ) delete sub; }

	virtual bool intersect( const ray& r, isect& i ) const;
	virtual bool intersect( const ray& r ) const;

	virtual const Material& getMaterial() const { return sub->getMaterial(); }
	virtual void setMaterial( Material *m ) { sub->setMaterial( m ); }

	virtual bool isTransformed() const { return true; }
	virtual SceneObject *compact();

protected:
	SceneObject	*sub;

	mat4 		xform;
	mat4		inverse;
	mat3 		normi;
};

class Scene
{
public:
	typedef list<Light*>::iterator 			liter;
	typedef list<Light*>::const_iterator 	cliter;

	typedef list<Geometry*>::iterator 		giter;
	typedef list<Geometry*>::const_iterator cgiter;

public:
	Scene() 
		: objects(), lights() {}
	virtual ~Scene();

	void add( Geometry* obj )
	{ objects.push_back( obj ); }
	void add( Light* light )
	{ lights.push_back( light ); }

	bool intersect( const ray& r, isect& i ) const;
	bool intersect( const ray& r ) const;
	bool intersect( const ray& r, double tmax ) const;

	cliter beginLights() const { return lights.begin(); }
	cliter endLights() const { return lights.end(); }

        Camera *getCamera() { return &camera; }
private:
    list<Geometry*> objects;
    list<Light*> lights;

    Camera camera;
};

#endif // __SCENE_H__
